// modified by Marc Mooser

// FS6IPC various utilities - Alessandro Antonini
// adapted from original code by Mr. Szofran
// NOTE: This unit is No OOP - I hate OOP so i try to avoid it when i can.
{$A-} // don't align



Unit IpcUtils;

Interface

Uses Fs6State, Windows, SysUtils;

// Define a structure for storing flight data.
Type tFlightData = Packed Record
    byClockHour,
    byClockMin,
    byClockSec  : Byte;
    dwPauseFlag,
    dwLatLo,                // latitude  (low dword)
    dwLatHi,                // latitude  (high dword)
    dwLonLo,                // longitude (low dword)
    dwLonHi,                // longitude (high dword)
    dwAltLo,                // altitude  (low dword)
    dwAltHi     : Dword;    // altitude  (high dword)
    dwTAS       : Dword;    // +02B4 true airspeed (m/s * 256)
    dwHeading   : dword;    // +05C8 Heading
End;

// Define a structure to hold data specific to each version of FS
// This structure helps ease compatibility with both FS95 and FS98.
Type fsInfo = Packed Record
    pszWindowClass  : String;   
    offset_byClockHour,
    offset_dwPauseFlag,
    offset_dwLatLo,
    offset_wGear1,
    offset_wGear2,
    offset_wGear3,
    offset_byLightsOn,
    offset_byStrobeOn,
    offset_dwTAS                : Dword; 
    offset_dwHeading,
    offset_byLandingLightOn     : Dword;
    wGearMask                   : Word;
End;

// structure, store some of most commons FS95 state VARs offset
Const FS95Info : fsInfo =
        (   pszWindowClass              : 'FS6MAIN';   
            offset_byClockHour          : $238;
            offset_dwPauseFlag          : $264;
            offset_dwLatLo              : $5A8;
            offset_wGear1               : $BC0;
            offset_wGear2               : $BC2;
            offset_wGear3               : $BC4;
            offset_byLightsOn           : $288;
            offset_byStrobeOn           : $289;
            offset_dwTAS                : $2B4;
            offset_dwHeading            : $5C8;
            offset_byLandingLightOn     : $298;
            wGearMask                   : $FFFF;
        );
    
// structure, store some of most commons FS98 state VARs offset
Const FS98Info : fsInfo =
        (   pszWindowClass              : 'FS98MAIN';   
            offset_byClockHour          : $238;
            offset_dwPauseFlag          : $264;
            offset_dwLatLo              : $560;
            offset_wGear1               : $BE8;
            offset_wGear2               : $BE8;
            offset_wGear3               : $BE8;
            offset_byLightsOn           : $280;
            offset_byStrobeOn           : $281;
            offset_dwTAS                : $2B8;
            offset_dwHeading            : $580;
            offset_byLandingLightOn     : $298;
            wGearMask                   : $3FFF;
        );


var FlightData : tFlightData;

// miscellaneous utils to make life easier
Function IpcGetLat( Clnt: FS6IPC_Client; 
                    Info: fsInfo; 
                    var sLat, sLon,
		    lat_string, lon_string : ShortString ): Boolean;
Function IpcGetAll( Clnt: FS6IPC_Client; Info: fsInfo;  var heading, TAS, alt, pause: longint  ): Boolean;
Function dwHeading2Deg( lat_long : LongInt ): Extended;



Implementation

// format heading values to return 0-360
Function dwHeading2Deg( lat_long : LongInt ): Extended;          

Var Lon : Extended;

Begin
    lon:=lat_long / 256.0 / ($01000000) * 360.0;
    If (lon>180.0) Then lon:=lon-360.0;
    If (lon<0.0) Then lon:=lon+360.0;
    Result := lon;
End; 


//// changes by MM //////////////////////////////////////////////////////////////////////////////////////////////////

Function LatStr( i : Extended): ShortString; 

var dd, mm, ss , mml: Extended;
    dds, mms, sss   : String;
    sign : Char;

Begin                         // i=9.76351666666666 
    dd := Trunc(i);           // Grade   (eg. 9)

    mm := (i-dd)*60;          // Minuten... (eg: 45.811)
    mml := Trunc(mm);         // Minuten (eg. 45)
    ss := (mm - mml) * 1000;  // Nachkommastellen der Minuten (eg: 0.811)

    Sign:='-';
    If dd>0.0 Then Sign := '+'
    Else if mml>0.0 Then Sign:='+'
         else if ss>0.0 Then Sign:='+';

    Str(Abs(dd):0:0,dds);     // create strings
    Str(Abs(mml):0:0,mms);
    Str(Abs(ss):0:0,sss);

    If Abs(dd)<10  Then dds:='0'+dds;
    If Abs(mml)<10 Then mms:='0'+mms;
    If Abs(ss)<10  Then sss:='0'+sss;
    If Abs(ss)<100 Then sss:='0'+sss;

    Result:=sign+dds+mms+'.'+sss;

    // result should be in this format:   (NMEA-compatible)    
    //  +70.00.999    -01.59.999 
    //  +01.00.999    +00.00.000 

End;


Function LonStr( i : Extended): ShortString; 

var dd, mm, ss , mml: Extended;
    dds, mms, sss   : String;
    sign : Char;

Begin                         // i=9.76351666666666 
    dd := Trunc(i);           // Grade   (eg. 9)
    mm := (i-dd)*60;          // Minuten... (eg: 45.811)
    mml := Trunc(mm);         // Minuten (eg. 45)
    ss := (mm - mml) * 1000;  // Nachkommastellen der Minuten (eg: 0.811)

    Sign:='-';
    If dd>0.0 Then Sign := '+'
    Else if mml>0.0 Then Sign:='+'
         else if ss>0.0 Then Sign:='+';

    Str(Abs(dd):0:0,dds);     // create strings
    Str(Abs(mml):0:0,mms);
    Str(Abs(ss):0:0,sss);

    If Abs(dd)<10  Then dds:='0'+dds;
    If Abs(dd)<100  Then dds:='0'+dds;    // special in LonStr
    If Abs(mml)<10 Then mms:='0'+mms;
    If Abs(ss)<10  Then sss:='0'+sss;
    If Abs(ss)<100 Then sss:='0'+sss;

    Result:=sign+dds+mms+'.'+sss;

    // result should be in this format:   (NMEA-compatible)  
    //  +170.00.000    +180.00.000
    //  +070.00.999    -001.59.999 
    //  +001.00.999    +000.00.000 

End;


Function human_readable( i : Extended): ShortString; 

var dd, mm, ss , mml: Extended;
    dds, mms, sss   : String;
    sign : Char;

Begin
    dd := Trunc(i);
    If dd>=0.0 Then Sign := '+' Else sign := '-';

    mm := (i-dd)*60;
    mml := Trunc(mm);
    ss := (mm - mml) * 60;

    Str(Abs(dd):0:0,dds);
    Str(Abs(mml):0:0,mms);
    Str(Abs(ss):0:2,sss);
    
    If Abs(mml)<10 Then mms:='0'+mms;
    If Abs(ss)<10 Then sss:='0'+sss;

    Result := dds+' deg '+mms+''''+sss+'"';

End;


/// **** IpcGetLat *** ///////////


Function IpcGetLat( Clnt: FS6IPC_Client; 
                    Info: fsInfo; 
                    var sLat, sLon,
		    lat_string, lon_string : ShortString): Boolean;
Var lat, lon : Extended;
                        
str_dwLatHi, str_dwLatLo, str_lat:ShortString; _dwLatHi, _dwLatLo:Extended;


Begin
    With Clnt Do Begin
        BeginRequest;       // start quering the IPC server              

        ReadStateData(Info.offset_dwLatLo, 6*sizeof(Dword), 
                      @flightdata.dwLatLo);
        If Not(SendRequest) Then Begin      // result failed
            sLat:='';    // clear eventual rubbish
            sLon:='';    
            Result:=False;  
            Exit;       // outta here
        End
        Else Begin

            With flightdata Do Begin        
	                                                                                  //dwLatHi = 32bit
                // LATITUDE
		// ========
		If (dwLatHi>0) Then lat:=dwLatHi   +    dwLatLo / 68719476736.0           // = 2^36
                               Else lat:=dwLatHi   -    dwLatLo / 68719476736.0;
	        
		if (dwLatHi>2147483648) Then lat:= lat - 4294967296.0;                    // 2^31 , 2^32
                
		lat:=lat*LatInt2Lat;                                                      // 45/5000875  =  8.99842527557677E-0006
                
        	sLat:=LatStr(lat);  // OK now we got it		
		lat_string := human_readable(lat);

		// output
		_dwLatHi := dwLatHi;
                _dwLatLo := dwLatLo;
		Str(Abs(_dwLatHi):0:0,str_dwLatHi);
		Str(Abs(_dwLatLo):0:0,str_dwLatLo);
		Str(Abs(lat):0:0,str_lat);
                //Writeln('****** ' + str_dwLatHi + ' -- ' + str_dwLatLo + ' ---- ' + str_lat);

		

		// LONGITUDE
		// =========
                If (dwLonHi>0) Then lon:=dwLonHi  +    dwLonLo / 68719476736.0
                               Else lon:=dwLonHi  -    dwLonLo / 68719476736.0;
                lon:=lon*LonInt2Lon;                                                       // 90/(256*4194304)  = 8.38190317153931E-0008   
                if lon > 180 then lon := lon - 360;
                sLon:=LonStr(lon);  // OK now we got it 
                lon_string := human_readable(lon);


            End;

        End;
    End;
    Result:=True;       // success
End;


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// gets all needed variables
// returns TRUE if successful, FALSE otherwise
Function IpcGetAll( Clnt: FS6IPC_Client; Info: fsInfo;  var heading, TAS, alt, pause: longint ): Boolean;

Begin
    With Clnt Do Begin
        BeginRequest;       // start quering the IPC server

        ReadStateData(Info.offset_dwPauseFlag, sizeof(Dword), 
                      @flightdata.dwPauseFlag);

        ReadStateData(Info.offset_dwLatLo, 6*sizeof(Dword), 
                      @flightdata.dwLatLo);
        // get clock
        ReadStateData(Info.offset_byClockHour, 3, @flightdata.byClockHour);
        // get Tas
        ReadStateData(Info.offset_dwTas, SizeOf(Dword),
                      @flightdata.dwTAS);
        // get Heading
        ReadStateData(Info.offset_dwHeading, SizeOf(Dword),
                      @flightdata.dwHeading);
        If Not(SendRequest) Then Begin      // result failed
            Result:=False;  
            Exit;       // outta here
        End
//!!        
        Else Begin
           heading := flightdata.dwHeading;
           TAS     := flightdata.dwTAS;
           alt     := flightdata.dwAltHi;
           pause   := flightdata.dwPauseFlag;
           
        End;

    End;
    Result:=True; // success

End;


Begin
    // init destination structure
    FillChar(FlightData,SizeOf(FlightData),#0);
End.
